import numpy as np

from tensorflow.keras.models import Model

import tensorflow as tf
from tensorflow.python.keras.utils.data_utils import Sequence
import pandas as pd
from sklearn.cluster import AgglomerativeClustering

import csv
import gc

config = tf.compat.v1.ConfigProto()
config.gpu_options.allow_growth = True  # dynamically grow the memory used on the GPU
config.log_device_placement = True  # to log device placement (on which device the operation ran)
sess = tf.compat.v1.Session(config=config)


class AugSequence(Sequence):
    def __init__(self, x_set, batch_size=32, augmentations=lambda image: image, layer_name='flatten'):
        self.x = x_set
        self.batch_size = batch_size
        self.augment = augmentations
        self.layer_name = layer_name

    def __len__(self):
        return int(np.ceil(len(self.x) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.x[idx * self.batch_size:(idx + 1) * self.batch_size]
        # batch_y = self.y[idx * self.batch_size:(idx + 1) * self.batch_size]

        return np.stack([self.augment(image=x)["image"] for x in batch_x], axis=0)


class TTABooster:
    def __init__(self, model, n_tta=30, k=10, batch_size=100, augmentations=[], reduction='ttabooster'):
        print("TTABoost ctor")
        self.model = model
        self.n_tta = n_tta
        self.k = k
        self.batch_size = batch_size
        self.augmentations = augmentations

    @staticmethod
    def top_k_accuracy(y_true, y_pred, k=1):
        '''From: https://github.com/chainer/chainer/issues/606

        Expects both y_true and y_pred to be one-hot encoded.
        '''
        argsorted_y = np.argsort(y_pred)[:, -k:]
        return np.any(argsorted_y.T == y_true.argmax(axis=1), axis=0).mean()

    @staticmethod
    def chunker(seq, size):
        return (seq[pos:pos + size] for pos in range(0, len(seq), size))

    def features_and_probs(self, x_chunck):
        layer_name = self.layer_name  # TODO: Extract the last layer name from the model
        intermediate_layer_model = Model(inputs=self.model.input,
                                         outputs=[self.model.get_layer(layer_name).output, self.model.output])
        valid_gen = AugSequence(x_set=x_chunck, batch_size=self.batch_size, augmentations=self.augmentations[0])
        # TODO: Make AugSequence work with multiple functions
        intermediate_output = intermediate_layer_model.predict(valid_gen, workers=1,
                                                               steps=len(x_chunck) * self.n_tta / self.batch_size
                                                               )
        embeddings, probs = intermediate_output[0], intermediate_output[1]
        del intermediate_output
        ## values ndarray will be of the shape (len(x_chuck),n_tta*n_classes) - for each sample, the predictions of all it's ttas
        sample_probs = probs[0:len(x_chunck),
                       :]  # Init the array with the first sample result (of all augmnetations)
        for i in range(self.n_tta - 1):  # Fill with rest of samples
            sample_probs = np.hstack([sample_probs, probs[(i + 1) * len(x_chunck):len(x_chunck) * (i + 2), :]])
        sample_features = embeddings[0:len(x_chunck), :]
        del probs
        gc.collect()
        for i in range(self.n_tta - 1):
            # gc.collect()
            sample_features = np.hstack(
                [sample_features, embeddings[(i + 1) * len(x_chunck):len(x_chunck) * (i + 2), :]])
        del embeddings
        gc.collect()
        return sample_features, sample_probs

    def predict(self, x, chunk_size=100):
        boost_ens = pd.DataFrame()
        all_ens = pd.DataFrame()
        random_ens = pd.DataFrame()
        
        for index, x_chunck in enumerate(TTABooster.chunker(x, chunk_size)):
            print("Started chunk", index, "of size", chunk_size)
            y_pred = self.model.predict(x_chunck, batch_size=self.batch_size, verbose=1)
            sample_features, sample_probs = self.features_and_probs(x_chunck)
            random_sample_probs = sample_probs[:, 0:sample_probs.shape[1] * self.k]

            for i in range(sample_features.shape[0]):
                AC = AgglomerativeClustering(n_clusters=self.k, linkage='average', affinity='cosine').fit(
                    np.stack([sample_features[i:i + 1, x:x + sample_features.shape[1]] for x in
                              range(0, sample_features.shape[1], sample_features.shape[1])],
                             axis=1).reshape(self.n_tta, sample_features.shape[1]))
                temp = pd.DataFrame(
                    np.stack(
                        [sample_probs[i:i + 1, x:x + sample_probs.shape[1]] for x in range(0, sample_probs.shape[1], sample_probs.shape[1])],
                        axis=1).reshape(self.n_tta, sample_probs.shape[1]))

                temp = temp.append(pd.DataFrame(y_pred[i, :]).T)
                AC.labels_ = np.hstack((AC.labels_, 999))
                temp['label'] = AC.labels_
                temp.drop_duplicates(subset='label', inplace=True)
                temp.drop('label', axis=1, inplace=True)
                boost_ens = boost_ens.append(pd.DataFrame(temp.mean(axis=0))[0])
                all_ens = all_ens.append(pd.DataFrame(temp.mean(axis=0))[0])
                if i % 1000 == 0:
                    print(i)

            for i in range(random_sample_probs.shape[0]):
                random_temp = pd.DataFrame(
                    np.stack([random_sample_probs[i:i + 1, x:x + sample_probs.shape[1]] for x in
                              range(0, random_sample_probs.shape[1], sample_probs.shape[1])],
                             axis=1).reshape(self.k, sample_probs.shape[1]))

                random_temp = random_temp.append(pd.DataFrame(y_pred[i, :]).T)
                random_ens = random_ens.append(pd.DataFrame(random_temp.mean(axis=0))[0])

        return boost_ens.values, all_ens.values, random_ens.values

    def benchmark_results(self, x, y):
        results = {}
        print("Predicting without tta...")
        y_pred = self.model.predict(x, batch_size=self.batch_size, verbose=1)
        no_tta_acc_top_1, no_tta_acc_top_5 = TTABooster.top_k_accuracy(y, y_pred, k=1), TTABooster.top_k_accuracy(y,
                                                                                                                  y_pred,
                                                                                                                  k=5)
        results['No TTA'] = {'top1': no_tta_acc_top_1, 'top5': no_tta_acc_top_5}

        print("Predicting with tta selection ...")
        y_pred_ttaboost, y_pred_tta_all, y_pred_random_tta = self.predict(x)
        tta_boost_acc_top_1, tta_boost_acc_top_5 = TTABooster.top_k_accuracy(y, y_pred_ttaboost,
                                                                             k=1), TTABooster.top_k_accuracy(y,
                                                                                                             y_pred_ttaboost,
                                                                                                             k=5)
        results['TTABoost'] = {'top1': tta_boost_acc_top_1, 'top5': tta_boost_acc_top_5}

        print("Predicting with random tta selection ...")
        random_tta_acc_top_1, random_tta_acc_top_5 = TTABooster.top_k_accuracy(y, y_pred_random_tta,
                                                                               k=1), TTABooster.top_k_accuracy(y,
                                                                                                               y_pred_random_tta,
                                                                                                               k=5)
        results['Random TTA'] = {'top1': random_tta_acc_top_1, 'top5': random_tta_acc_top_5}

        print("Predicting with all tta...")
        all_tta_acc_top_1, all_tta_acc_top_5 = TTABooster.top_k_accuracy(y, y_pred_tta_all,
                                                                         k=1), TTABooster.top_k_accuracy(y,
                                                                                                         y_pred_tta_all,
                                                                                                         k=5)
        results['All TTA'] = {'top1': all_tta_acc_top_1, 'top5': all_tta_acc_top_5}

        self.save_benchmark_result(results)

    def save_benchmark_result(self, results, results_file='benchmark-results.csv'):
        with open(results_file, 'a+', newline='') as csvfile:
            fieldnames = ['TTA_method', 'top1', 'top5', 'Notes']
            res_writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            res_writer.writeheader()

            for method_name, accuracies in results.items():
                res_writer.writerow({'TTA_method': method_name,
                                     'top1': accuracies['top1'],
                                     'top5': accuracies['top5'],
                                     'Notes': ''})



from ttabooster.datasets import cifar10_validation_set
from ttabooster.TTABoost import TTABooster
from tensorflow.keras.models import load_model

from albumentations import (
    Compose, RandomSizedCrop
)

model = load_model('../saved_models/Keras_cifar10_200epc_resnet110v2_for_Kaduri.h5')  # Load any keras model
x_test, y_test = cifar10_validation_set(limit=200)  # Load your validation set

AUGMENTATIONS_TEST = Compose([
    RandomSizedCrop((28, 28), 32, 32, w2h_ratio=1.0, interpolation=1, always_apply=False, p=1.0),
])

booster = TTABooster(model, batch_size=100, augmentations=AUGMENTATIONS_TEST)
booster.benchmark_results(x_test, y_test)



from __future__ import print_function
import tensorflow
from tensorflow.keras.layers import Dense, Conv2D, BatchNormalization, Activation
from tensorflow.keras.layers import AveragePooling2D, Input, Flatten
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.callbacks import ModelCheckpoint, LearningRateScheduler
from tensorflow.keras.callbacks import ReduceLROnPlateau
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.regularizers import l2
from tensorflow.keras import backend as K
from tensorflow.keras.models import Model
from tensorflow.keras.datasets import cifar10
import numpy as np
import os

# Training parameters
batch_size = 32  # orig paper trained all networks with batch_size=128
epochs = 200
data_augmentation = True
num_classes = 10

# Subtracting pixel mean improves accuracy
subtract_pixel_mean = True

# Model parameter
# ----------------------------------------------------------------------------
#           |      | 200-epoch | Orig Paper| 200-epoch | Orig Paper| sec/epoch
# Model     |  n   | ResNet v1 | ResNet v1 | ResNet v2 | ResNet v2 | GTX1080Ti
#           |v1(v2)| %Accuracy | %Accuracy | %Accuracy | %Accuracy | v1 (v2)
# ----------------------------------------------------------------------------
# ResNet20  | 3 (2)| 92.16     | 91.25     | -----     | -----     | 35 (---)
# ResNet32  | 5(NA)| 92.46     | 92.49     | NA        | NA        | 50 ( NA)
# ResNet44  | 7(NA)| 92.50     | 92.83     | NA        | NA        | 70 ( NA)
# ResNet56  | 9 (6)| 92.71     | 93.03     | 93.01     | NA        | 90 (100)
# ResNet110 |18(12)| 92.65     | 93.39+-.16| 93.15     | 93.63     | 165(180)
# ResNet164 |27(18)| -----     | 94.07     | -----     | 94.54     | ---(---)
# ResNet1001| (111)| -----     | 92.39     | -----     | 95.08+-.14| ---(---)
# ---------------------------------------------------------------------------
n = 12

# Model version
# Orig paper: version = 1 (ResNet v1), Improved ResNet: version = 2 (ResNet v2)
version = 2

# Computed depth from supplied model parameter n
if version == 1:
    depth = n * 6 + 2
elif version == 2:
    depth = n * 9 + 2

# Model name, depth and version
model_type = 'ResNet%dv%d' % (depth, version)

# Load the CIFAR10 data.
(x_train, y_train), (x_test, y_test) = cifar10.load_data()

# Input image dimensions.
input_shape = x_train.shape[1:]

# Normalize data.
x_train = x_train.astype('float32') / 255
x_test = x_test.astype('float32') / 255

# If subtract pixel mean is enabled
if subtract_pixel_mean:
    x_train_mean = np.mean(x_train, axis=0)
    x_train -= x_train_mean
    x_test -= x_train_mean

print('x_train shape:', x_train.shape)
print(x_train.shape[0], 'train samples')
print(x_test.shape[0], 'test samples')
print('y_train shape:', y_train.shape)

# Convert class vectors to binary class matrices.
y_train = tensorflow.keras.utils.to_categorical(y_train, num_classes)
y_test = tensorflow.keras.utils.to_categorical(y_test, num_classes)


def lr_schedule(epoch):
    """Learning Rate Schedule

    Learning rate is scheduled to be reduced after 80, 120, 160, 180 epochs.
    Called automatically every epoch as part of callbacks during training.

    # Arguments
        epoch (int): The number of epochs

    # Returns
        lr (float32): learning rate
    """
    lr = 1e-3
    if epoch > 180:
        lr *= 0.5e-3
    elif epoch > 160:
        lr *= 1e-3
    elif epoch > 120:
        lr *= 1e-2
    elif epoch > 80:
        lr *= 1e-1
    print('Learning rate: ', lr)
    return lr


def resnet_layer(inputs,
                 num_filters=16,
                 kernel_size=3,
                 strides=1,
                 activation='relu',
                 batch_normalization=True,
                 conv_first=True):
    """2D Convolution-Batch Normalization-Activation stack builder

    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)

    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv2D(num_filters,
                  kernel_size=kernel_size,
                  strides=strides,
                  padding='same',
                  kernel_initializer='he_normal',
                  kernel_regularizer=l2(1e-4))

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x


def resnet_v1(input_shape, depth, num_classes=10):
    """ResNet Version 1 Model builder [a]

    Stacks of 2 x (3 x 3) Conv2D-BN-ReLU
    Last ReLU is after the shortcut connection.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filters is
    doubled. Within each stage, the layers have the same number filters and the
    same number of filters.
    Features maps sizes:
    stage 0: 32x32, 16
    stage 1: 16x16, 32
    stage 2:  8x8,  64
    The Number of parameters is approx the same as Table 6 of [a]:
    ResNet20 0.27M
    ResNet32 0.46M
    ResNet44 0.66M
    ResNet56 0.85M
    ResNet110 1.7M

    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)

    # Returns
        model (Model): Keras model instance
    """
    if (depth - 2) % 6 != 0:
        raise ValueError('depth should be 6n+2 (eg 20, 32, 44 in [a])')
    # Start model definition.
    num_filters = 16
    num_res_blocks = int((depth - 2) / 6)

    inputs = Input(shape=input_shape)
    x = resnet_layer(inputs=inputs)
    # Instantiate the stack of residual units
    for stack in range(3):
        for res_block in range(num_res_blocks):
            strides = 1
            if stack > 0 and res_block == 0:  # first layer but not first stack
                strides = 2  # downsample
            y = resnet_layer(inputs=x,
                             num_filters=num_filters,
                             strides=strides)
            y = resnet_layer(inputs=y,
                             num_filters=num_filters,
                             activation=None)
            if stack > 0 and res_block == 0:  # first layer but not first stack
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_layer(inputs=x,
                                 num_filters=num_filters,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
            x = tensorflow.keras.layers.add([x, y])
            x = Activation('relu')(x)
        num_filters *= 2

    # Add classifier on top.
    # v1 does not use BN after last shortcut connection-ReLU
    x = AveragePooling2D(pool_size=8)(x)
    y = Flatten()(x)
    outputs = Dense(num_classes,
                    activation='softmax',
                    kernel_initializer='he_normal')(y)

    # Instantiate model.
    model = Model(inputs=inputs, outputs=outputs)
    return model


def resnet_v2(input_shape, depth, num_classes=10):
    """ResNet Version 2 Model builder [b]

    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256

    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)

    # Returns
        model (Model): Keras model instance
    """
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = 16
    num_res_blocks = int((depth - 2) / 9)

    inputs = Input(shape=input_shape)
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_layer(inputs=inputs,
                     num_filters=num_filters_in,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(3):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # first layer but not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_layer(inputs=x,
                             num_filters=num_filters_in,
                             kernel_size=1,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization,
                             conv_first=False)
            y = resnet_layer(inputs=y,
                             num_filters=num_filters_in,
                             conv_first=False)
            y = resnet_layer(inputs=y,
                             num_filters=num_filters_out,
                             kernel_size=1,
                             conv_first=False)
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_layer(inputs=x,
                                 num_filters=num_filters_out,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
            x = tensorflow.keras.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = AveragePooling2D(pool_size=8)(x)
    y = Flatten()(x)
    outputs = Dense(num_classes,
                    activation='softmax',
                    kernel_initializer='he_normal')(y)

    # Instantiate model.
    model = Model(inputs=inputs, outputs=outputs)
    return model


def cifar_model():
    if version == 2:
        model = resnet_v2(input_shape=input_shape, depth=depth)
    else:
        model = resnet_v1(input_shape=input_shape, depth=depth)

    model.compile(loss='categorical_crossentropy',
                  optimizer=Adam(learning_rate=lr_schedule(0)),
                  metrics=['accuracy'])
    model.summary()
    print(model_type)
    return model

# # Prepare model model saving directory.
# save_dir = os.path.join(os.getcwd(), 'saved_models')
# model_name = 'cifar10_%s_model.{epoch:03d}.h5' % model_type
# if not os.path.isdir(save_dir):
#     os.makedirs(save_dir)
# filepath = os.path.join(save_dir, model_name)
#
# # Prepare callbacks for model saving and for learning rate adjustment.
# checkpoint = ModelCheckpoint(filepath=filepath,
#                              monitor='val_acc',
#                              verbose=1,
#                              save_best_only=True)
#
# lr_scheduler = LearningRateScheduler(lr_schedule)
#
# lr_reducer = ReduceLROnPlateau(factor=np.sqrt(0.1),
#                                cooldown=0,
#                                patience=5,
#                                min_lr=0.5e-6)
#
# callbacks = [checkpoint, lr_reducer, lr_scheduler]



from setuptools import setup

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='ttabooster',
      version='0.14',
      url='https://github.com/OmriKaduri/ttabooster',
      license='MIT',
      author='Seffi Cohen, Omri Kaduri',
      author_email='Kaduriomri@gmail.com',
      description='Boost pretrained models with test time augmentation selection',
      packages=['ttabooster'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      download_url='https://github.com/OmriKaduri/ttabooster/archive/0.14.zip',
      zip_safe=False,
      install_requires=[  # I get to this in a second
          'numpy>=1.17.4',
          'pandas>=0.25.3',
          'tensorflow>=2.0.0',
          'scikit-learn>=0.22'],
      )



from tensorflow.keras.datasets import cifar10
import numpy as np
import tensorflow as tf


def cifar10_validation_set(subtract_pixel_mean=True, limit=None):
    (x_train, y_train), (x_test, y_test) = cifar10.load_data()
    num_classes = 10

    # Normalize data.
    x_train = x_train.astype('float32') / 255
    x_test = x_test.astype('float32') / 255

    # If subtract pixel mean is enabled
    if subtract_pixel_mean:
        x_train_mean = np.mean(x_train, axis=0)
        x_train -= x_train_mean
        x_test -= x_train_mean

    y_test = tf.keras.utils.to_categorical(y_test, num_classes)
    if limit:
        return x_test[:limit], y_test[:limit]
    return x_test, y_test
